import { serve } from "./deps/std.ts";

import {
	ConnectAnswer,
	ConnectOffer,
	ConnectReject,
	ICECandidate,
	PearRequest,
	PearResponse,
	PeerID,
} from "../common/types.ts";
import { BiMap } from "../common/bimap.ts";

type Peer = WebSocket;

const peers: BiMap<Peer, PeerID> = new BiMap();
const getPeerID = (peer: Peer) => peers.getByA(peer);

const generateID = <T extends string>(): T => crypto.randomUUID() as T;
const generatePeerID = generateID<PeerID>;

const send = (peer: Peer, response: PearResponse) => {
	peer.send(
		JSON.stringify(response, (_key, value) => {
			if (value instanceof Set) {
				return [...value.values()];
			}

			return value;
		}),
	);
};

const getOtherPeer = (thisPeerID: PeerID, otherPeerID: PeerID) => {
	if (thisPeerID === otherPeerID) {
		return "Peer not found";
	}

	const peer = peers.getByB(otherPeerID);
	if (!peer) {
		return "Peer not found";
	}

	return peer;
};

const connectOffer = (peer: Peer, peerID: PeerID, msg: ConnectOffer) => {
	const otherPeer = getOtherPeer(peerID, msg.peerID);
	if (typeof otherPeer === "string") {
		send(peer, { error: otherPeer });
		return;
	}

	send(otherPeer, { kind: "connectOffer", peerID, offer: msg.offer });
};

const connectAnswer = (peer: Peer, peerID: PeerID, msg: ConnectAnswer) => {
	const otherPeer = getOtherPeer(peerID, msg.peerID);
	if (typeof otherPeer === "string") {
		send(peer, { error: otherPeer });
		return;
	}

	send(otherPeer, { kind: "connectAnswer", peerID, answer: msg.answer });
};

const iceCandidate = (peer: Peer, peerID: PeerID, msg: ICECandidate) => {
	const otherPeer = getOtherPeer(peerID, msg.peerID);
	if (typeof otherPeer === "string") {
		send(peer, { error: otherPeer });
		return;
	}

	send(otherPeer, { kind: "iceCandidate", peerID, candidate: msg.candidate });
};

const connectReject = (peer: Peer, peerID: PeerID, msg: ConnectReject) => {
	const otherPeer = getOtherPeer(peerID, msg.peerID);
	if (typeof otherPeer === "string") {
		send(peer, { error: otherPeer });
		return;
	}

	send(otherPeer, { kind: "connectReject", peerID });
};

const HANDLERS = {
	connectAnswer,
	connectOffer,
	connectReject,
	iceCandidate,
} satisfies {
	[K in PearRequest["kind"]]: (
		thisPeer: Peer,
		thisPeerID: PeerID,
		data: PearRequest & { kind: K },
	) => void;
};
const handleSocketMessage = function (this: WebSocket, ev: MessageEvent) {
	const data: PearRequest = JSON.parse(ev.data);
	console.debug(data);

	const peerID = getPeerID(this);
	if (!peerID) {
		console.error(`Message from unregistered peer: ${data}`);
		return;
	}

	const handler = HANDLERS[data.kind];
	handler(this, peerID, data as never);
};

const cleanupPeer = function (this: WebSocket) {
	const peerID = getPeerID(this);
	if (!peerID) {
		return;
	}

	console.log(`Cleaning up peer ${peerID}`);

	peers.removeByB(peerID);
};

const handleReq = (req: Request) => {
	if (req.headers.get("upgrade") != "websocket") {
		return new Response(null, { status: 501 });
	}

	const { socket, response } = Deno.upgradeWebSocket(req);

	const peerID = generatePeerID();
	peers.set(socket, peerID);

	socket.onopen = () => {
		send(socket, { kind: "registered", peerID });
	};
	socket.onmessage = handleSocketMessage;
	socket.onclose = cleanupPeer;

	console.log(`Peer ${peerID} connected`);

	return response;
};

const port = +(Deno.env.get("PEAR_SERVER_PORT") ?? 8382);
await serve(handleReq, {
	port,
	onError: (...args) => {
		console.debug(args);

		return new Response(null, { status: 500 });
	},
});
