import { swel } from "./deps/swel.ts";
import {
	map as createMap,
	tileLayer as createTileLayer,
	marker as createMarker,
	Marker,
} from "./deps/leaflet.ts";

import { AcceptDecision, initNetworking } from "./networking.ts";
import { PeerID } from "../common/types.ts";
import { showModal } from "./ui.ts";
import { WalkyError } from "./error.ts";

type LatLng = {
	lat: number;
	lng: number;
};
type Peer = {
	marker: Marker | null;
	dataChannel: RTCDataChannel;
	listItemEl: HTMLElement;
	statusEl: HTMLElement;
};
const init = async () => {
	const peers = new Map<PeerID, Peer>();
	const onPeerConnecting = async (
		peerID: PeerID,
		rtc: RTCPeerConnection,
		connected: Promise<RTCDataChannel>,
	) => {
		const statusEl = swel("div", { className: "peerStatus" }, "Connecting");
		const listItemEl = swel("div", { className: "peer" }, [
			swel("div", { className: "peerID" }, peerID),
			statusEl,
		]);
		peersList.appendChild(listItemEl);

		const dataChannel = await connected;
		statusEl.textContent = "Connected";

		const peer = {
			marker: null,
			dataChannel,
			listItemEl,
			statusEl,
		} as Peer;
		peers.set(peerID, peer);

		dataChannel.addEventListener("message", (msg) => {
			const latLng: LatLng = JSON.parse(msg.data);

			if (!peer.marker) {
				peer.marker = createMarker(latLng, {
					title: `${peerID}`,
					autoPan: true,
				}).bindPopup(`${peerID}`);
				peer.marker.addTo(map);
			} else {
				peer.marker.setLatLng(latLng);
			}
		});

		dataChannel.addEventListener("open", () => {
			if (lastPosition) {
				dataChannel.send(JSON.stringify(lastPosition));
			}
		});

		dataChannel.addEventListener("close", () => {
			peer.statusEl.textContent = "Disconnected";
		});
	};

	const onPeerConnectionRequest = (peerID: PeerID) => {
		return new Promise<AcceptDecision>((r) => {
			let secondsLeft = 30;
			const interval = setInterval(() => {
				secondsLeft--;
				if (secondsLeft <= 0) {
					r(false);
					closeModal();
					clearInterval(interval);
				}

				timeLeft.textContent = `${secondsLeft}s`;
			}, 1000);
			const timeLeft = swel("div", `${secondsLeft}s`);

			const onClose = () => {
				clearInterval(interval);
			};

			const closeModal = showModal({
				content: [
					`Accept connection from ${peerID}?`,
					timeLeft,
					swel(
						"button",
						{
							on: {
								click: () => {
									r(true);
									closeModal(false);
								},
							},
						},
						"Accept",
					),
					swel(
						"button",
						{
							on: {
								click: () => {
									r(false);
									closeModal(false);
								},
							},
						},
						"Reject",
					),
				],
				onClose: () => {
					r(false);
					onClose();
				},
			});
		});
	};

	const onConnectionRejected = (peerID: PeerID) => {
		showModal({ content: `${peerID} rejected your connection request` });
	};

	const net = await initNetworking({
		onPeerConnecting,
		onPeerConnectionRequest,
		onConnectionRejected,
	});

	const peerIDInput = swel("input", { placeholder: "Peer ID" });
	const connectButton = swel(
		"button",
		{
			on: {
				click: () => {
					net.connectToPeer(peerIDInput.value as PeerID);
					peerIDInput.value = "";
				},
			},
		},
		"Connect",
	);

	const peersList = swel("div", { id: "peers" });
	const lobby = swel("div", { id: "lobby" }, [
		peersList,
		swel("div", { className: "controls" }, [
			swel("div", { id: "peerID" }, [
				`Your ID: ${net.peerID}`,
				swel(
					"button",
					{
						on: {
							click: () => {
								navigator.clipboard.writeText(`${net.peerID}`);
							},
						},
					},
					"📋",
				),
			]),
			swel("div", [peerIDInput, connectButton]),
		]),
	]);

	const mapContainer = swel("div", { id: "map" });

	const main = document.getElementsByTagName("main")[0];
	main.replaceChildren(mapContainer, lobby);

	const map = createMap(mapContainer);
	map.setView([37.8, -96], 4);

	const tileLayer = createTileLayer(
		"https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png",
		{
			maxZoom: 19,
			attribution:
				'&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
		},
	);
	tileLayer.addTo(map);

	let myMarker: Marker | null = null;
	let lastPosition: LatLng | null = null;
	navigator.geolocation.watchPosition((p) => {
		const latLng = {
			lat: p.coords.latitude,
			lng: p.coords.longitude,
		};
		if (!myMarker) {
			myMarker = createMarker(latLng, { title: "Me", autoPan: true }).bindPopup(
				"Me",
			);
			myMarker.addTo(map);
		} else {
			myMarker.setLatLng(latLng);
		}

		for (const [_, peer] of peers) {
			peer.dataChannel.send(JSON.stringify(latLng));
		}

		lastPosition = latLng;
	});
};

(window as unknown as { init: typeof init })["init"] = init;
addEventListener("error", (ev) => {
	if (ev.error instanceof WalkyError) {
		showModal({ content: `An error occured: ${ev.error}` });
	}
});
