import {
	ConnectAnswer,
	PearError,
	PearRequest,
	PearResponse,
	PeerID,
} from "../common/types.ts";

import { pearServerSocketAddress, rtcConfiguration } from "../ui.config.ts";
import { WalkyError } from "./error.ts";

export type AcceptDecision = true | false;

export const initNetworking = async (p: {
	onPeerConnectionRequest: (peerID: PeerID) => Promise<AcceptDecision>;
	onPeerConnecting: (
		peerID: PeerID,
		rtc: RTCPeerConnection,
		connected: Promise<RTCDataChannel>,
	) => void;
	onConnectionRejected: (peerID: PeerID) => void;
}) => {
	const socket = new WebSocket(pearServerSocketAddress);
	const pendingAnswers = new Map() as Map<
		PeerID,
		(r: ConnectAnswer["answer"]) => void
	>;
	const connections = new Map() as Map<PeerID, RTCPeerConnection>;

	type LR = Exclude<PearResponse, PearError>;
	const on: {
		[K in LR["kind"]]: ((k: LR & { kind: K }) => void) | null;
	} = {
		registered: null,
		connectAnswer: null,
		connectOffer: null,
		iceCandidate: null,
		connectReject: null,
	};

	socket.addEventListener("message", (ev) => {
		const data: PearResponse = JSON.parse(ev.data);
		console.debug(data);

		if ("error" in data) {
			throw new WalkyError(data.error);
		}

		const handler = on[data.kind];
		if (handler) {
			handler(data as never);
		}
	});

	const myPeerID: PeerID = await new Promise((r) => {
		on.registered = (msg) => {
			on.registered = null;
			r(msg.peerID);
		};
	});

	const send = (msg: PearRequest) => {
		const enc = JSON.stringify(msg);
		socket.send(enc);
	};

	const sendConnectOffer = (
		peerID: PeerID,
		offer: RTCSessionDescriptionInit,
	) => {
		const answer: Promise<RTCSessionDescriptionInit> = new Promise((r) => {
			pendingAnswers.set(peerID, r);
		});

		send({ kind: "connectOffer", offer, peerID });

		return answer;
	};

	on.connectAnswer = (msg) => {
		const pa = pendingAnswers.get(msg.peerID);
		if (!pa) {
			return;
		}

		pendingAnswers.delete(msg.peerID);
		pa(msg.answer);
	};

	const rtcPeerConnection = () => {
		const rtc = new RTCPeerConnection(rtcConfiguration);
		const connected = new Promise((r) => {
			rtc.addEventListener("iceconnectionstatechange", (_ev) => {
				console.debug(rtc.iceConnectionState);

				if (rtc.iceConnectionState == "connected") {
					r(undefined);
				}
			});
		});

		return { rtc, connected };
	};

	on.connectOffer = async (msg) => {
		const accept = await p.onPeerConnectionRequest(msg.peerID);
		if (!accept) {
			send({ kind: "connectReject", peerID: msg.peerID });
			return;
		}

		const { rtc, connected } = rtcPeerConnection();
		connections.set(msg.peerID, rtc);

		const dataChannelP: Promise<RTCDataChannel> = new Promise((r) => {
			rtc.ondatachannel = (ev) => {
				rtc.ondatachannel = null;

				r(ev.channel);
			};
		});
		await rtc.setRemoteDescription(msg.offer);

		const answer = await rtc.createAnswer();
		await rtc.setLocalDescription(answer);

		send({ kind: "connectAnswer", answer, peerID: msg.peerID });
		rtc.addEventListener("icecandidate", (ev) => {
			if (ev.candidate) {
				send({
					kind: "iceCandidate",
					candidate: ev.candidate,
					peerID: msg.peerID,
				});
			}
		});

		const connectedP = (async () => {
			const dataChannel = await dataChannelP;
			await connected;

			return dataChannel;
		})();

		p.onPeerConnecting(msg.peerID, rtc, connectedP);
	};

	const connectToPeer = async (peerID: PeerID) => {
		const { rtc, connected } = rtcPeerConnection();
		connections.set(peerID, rtc);

		const dataChannel = rtc.createDataChannel("data");

		const offer = await rtc.createOffer();
		await rtc.setLocalDescription(offer);

		const answer = await sendConnectOffer(peerID, offer);
		rtc.addEventListener("icecandidate", (ev) => {
			if (ev.candidate) {
				send({ kind: "iceCandidate", candidate: ev.candidate, peerID });
			}
		});
		await rtc.setRemoteDescription(answer);

		const connectedP = (async () => {
			await connected;

			return dataChannel;
		})();

		p.onPeerConnecting(peerID, rtc, connectedP);
	};

	on.iceCandidate = (msg) => {
		const rtc = connections.get(msg.peerID);
		if (!rtc) {
			return;
		}

		rtc.addIceCandidate(msg.candidate);
	};

	on.connectReject = (msg) => {
		p.onConnectionRejected(msg.peerID);
	};

	return {
		connectToPeer,
		peerID: myPeerID,
	};
};
